Build Packages / build:rpm (rocky9_sls9) (push) Successful in 18m57s
Build Packages / Unit tests (push) Skipped
Build Packages / build:windows:nocuda (push) Successful in 16m55s
Build Packages / build:windows:cuda (push) Successful in 18m48s
Build Packages / build:viewer-tgz:cpu (push) Successful in 13m10s
Build Packages / build:viewer-tgz:cuda (push) Successful in 14m45s
Build Packages / build:rpm (rocky8_nocuda) (push) Successful in 22m23s
Build Packages / build:rpm (rocky9_nocuda) (push) Successful in 20m12s
Build Packages / build:rpm (ubuntu2204_nocuda) (push) Successful in 23m7s
Build Packages / build:rpm (ubuntu2404_nocuda) (push) Successful in 20m43s
Build Packages / build:rpm (rocky8_sls9) (push) Successful in 23m9s
Build Packages / XDS test (durin plugin) (push) Successful in 12m26s
Build Packages / build:rpm (rocky9) (push) Successful in 24m58s
Build Packages / Generate python client (push) Successful in 50s
Build Packages / build:rpm (ubuntu2404) (push) Successful in 23m20s
Build Packages / Create release (push) Skipped
Build Packages / XDS test (JFJoch plugin) (push) Successful in 12m37s
Build Packages / build:rpm (rocky8) (push) Successful in 27m58s
Build Packages / build:rpm (ubuntu2204) (push) Successful in 25m38s
Build Packages / Build documentation (push) Successful in 59s
Build Packages / DIALS test (push) Successful in 23m16s
Build Packages / XDS test (neggia plugin) (push) Successful in 6m38s
**Files written by Jungfraujoch now import correctly in DIALS, XDS and pyFAI.** A tilted detector, a grid scan, a still recorded at a goniometer position, and saturated or unreadable pixels were each described in a way that a third-party program acted on wrongly. If you process Jungfraujoch data outside Jungfraujoch, prefer this release to any earlier one. * HDF5: the detector tilt (`rot1`/`rot2`/`rot3`) is exported correctly in the NXmx transformation chain; untilted geometries are unaffected. * HDF5: a still recorded at a goniometer position is no longer read back as a single image, and a grid scan records a stationary spindle so a program that requires a rotation axis can open it. * HDF5: the sample transformation chain is written in mounting order, with a Smargon head position told apart from the spindle, one entry per image, `module_offset` as a float unit vector, and `offset_units` on every offset. * HDF5: saturated, underloaded and unreadable pixels are described so a downstream program masks them - `saturation_value`, `underload_value`, `error_value` and `bit_depth_readout` are written correctly, and a data file missing next to a VDS master reads as the error marker rather than as zero counts. * HDF5: the rotation axis is read back under whatever name it carries, and `mirror_y` records whether the assembled image is mirrored in Y relative to the detector's raw readout. * A grid scan and a goniometer axis can both be set; they are no longer alternatives. * `images_per_file` is chosen from the acquisition when it is not given: a rotation sweep of at most 20000 images goes into a single data file, a grid scan splits on whole fast-axis rows, and stills and serial keep 1000. * The writer refuses a stream whose start message declares a different pixel format than its images carry, and a DECTRIS detector sending signed images is no longer declared unsigned. * The image stream can carry the sample transformation chain (`transformations`, in the END message); a producer that does not send it gets the same chain built by the writer. * rugnux: fixing the space group with `-S` no longer prevents the lattice from being found - a lattice indexed in a different setting is reindexed into that group's own setting, and a run whose crystal does not have that group's lattice stops and names the cell it indexed as, rather than reporting statistics that cannot describe it. * rugnux: the per-image resolution estimate now predicts the resolution the merged data reach rather than the highest-resolution spot found, and is reported as `SPOT_RESOLUTION_ESTIMATE`. * rugnux: two runs of the same command on the same images produce the same merged intensities; the azimuthal profile written alongside them is not yet reproducible in the same way. * rugnux: the offline lattice refinement is bounded by iterations rather than by a wall clock, so a loaded machine can no longer refine to a different lattice; a live acquisition keeps its real-time bound. * rugnux: the detector-frame modulation correction is fitted on a grid spanning the detector, so whether it is applied no longer depends on how far integration reached. * rugnux: the geometry pre-pass no longer writes `<prefix>_01.mtz`, `_01.cif`, `_01.hkl` and `_01_image.dat`; the refined second pass writes those files under `<prefix>`, and that is the result to use. * rugnux: `_process.h5` describes the pixel format of the images it links to, and is written on a thread of its own. * rugnux: the detector geometry is also logged in XDS's convention (`ORGX`/`ORGY`, detector axis vectors, rotation axis), so it can be compared with an XDS refinement. * rugnux: an image integrated in pyFAI through the `.poni` file written by `--mode calibration` comes out with the correct azimuth, and the file declares pyFAI's `orientation`, which needs pyFAI 2024.01 or newer. Radial integration is unchanged. * rugnux: a rotation run is substantially faster throughout - beam-stop detection, first-pass indexing, geometry refinement, integration, scaling and merging - and observations outside the scaling resolution range are dropped as they are ingested. The refined geometry, the space group chosen and the merged statistics are unchanged. * Faster spot finding and indexing, on the broker as well as in rugnux; the spots found and the lattices indexed are unchanged. * A run reserves substantially less GPU memory: nothing is allocated for buffers that are never read, and a worker builds only the engines it uses. * rugnux: with `-N` left at its default the per-image loop of `--mode mx` uses at most 16 workers per GPU, rather than one per hardware thread; an explicit `-N` is obeyed as given. * CUDA 12 builds now contain device code for Volta, so the RHEL 8 packages and the portable Linux `.tgz` run on a V100; the CUDA 13 artefacts (RHEL 9, Ubuntu, Windows) remain Turing and newer. * The build resolves a single Eigen for the whole project, and refuses to configure if Ceres picks up a different one; a build that mixed two Eigen versions was undefined behaviour and crashed at -O2. * Documentation: a security page, and the supported GPU generations and minimum NVIDIA driver version of every released artefact. **Breaking change to OpenAPI** - regenerate the client (`jfjoch-client` 1.0.0-rc.162, `frontend/src/client`): * `dataset_settings.images_per_file` is no longer `default: 1000` and no longer accepts `0`; it is optional, and its minimum is 1. A client sending `0` (previously "one file for the whole run") is now rejected - omit the field instead, which for a rotation sweep gives the same single file. * `file_writer_format` now defaults to `NXmxVDS`, matching the server's own default and the layout recommended for DIALS, XDS and CrystFEL. A generated client that fills in schema defaults and does not set the format explicitly will write VDS masters where it previously wrote legacy ones; set `NXmxLegacy` explicitly to keep them. --------- Co-authored-by: jungfrau <jungfrau@mx-aare-test.psi.ch> Reviewed-on: #72 Co-authored-by: Filip Leonarski <filip.leonarski@psi.ch>
192 lines
8.4 KiB
Plaintext
192 lines
8.4 KiB
Plaintext
// SPDX-FileCopyrightText: 2026 Filip Leonarski, Paul Scherrer Institute <filip.leonarski@psi.ch>
|
|
// SPDX-License-Identifier: GPL-3.0-only
|
|
|
|
#include "ShadowAccumulatorGPU.h"
|
|
|
|
#include <limits>
|
|
#include <type_traits>
|
|
|
|
#include "../../common/CUDAWrapper.h"
|
|
#include "../../common/JFJochException.h"
|
|
|
|
inline void cuda_err(cudaError_t val) {
|
|
if (val != cudaSuccess)
|
|
throw JFJochException(JFJochExceptionCategory::GPUCUDAError, cudaGetErrorString(val));
|
|
}
|
|
|
|
namespace {
|
|
|
|
// One frame folded into the projection. The sentinel test is predicated rather than branched: a
|
|
// warp straddling a module gap then costs the same as one that does not, against a per-frame memory
|
|
// bill of several hundred megabytes.
|
|
//
|
|
// The update is written exactly as the host writes it, including the "first value wins whatever it
|
|
// is" rule for the maximum - a pixel that has never been counted holds 0, which a genuinely negative
|
|
// maximum has to be allowed to replace. Each pixel is owned by one thread and the frames are
|
|
// separate launches on one stream, so there is no race and no atomic.
|
|
template <class T>
|
|
__global__ void accumulate_kernel(const T *__restrict__ frames, int nframes, size_t frame_stride,
|
|
int64_t *__restrict__ max_value, int64_t *__restrict__ sum_value,
|
|
uint32_t *__restrict__ valid_count, size_t npixels, T masked) {
|
|
for (size_t i = blockIdx.x * static_cast<size_t>(blockDim.x) + threadIdx.x; i < npixels;
|
|
i += static_cast<size_t>(blockDim.x) * gridDim.x) {
|
|
int64_t mx = max_value[i];
|
|
int64_t sm = sum_value[i];
|
|
uint32_t c = valid_count[i];
|
|
for (int k = 0; k < nframes; k++) {
|
|
const T v = frames[k * frame_stride + i];
|
|
if (v == masked)
|
|
continue;
|
|
const int64_t vi = static_cast<int64_t>(v);
|
|
if (c == 0 || vi > mx)
|
|
mx = vi;
|
|
sm += vi;
|
|
c++;
|
|
}
|
|
max_value[i] = mx;
|
|
sum_value[i] = sm;
|
|
valid_count[i] = c;
|
|
}
|
|
}
|
|
|
|
template <class T>
|
|
void launch(const uint8_t *raw, int nframes, size_t frame_bytes, int64_t *max_value,
|
|
int64_t *sum_value, uint32_t *valid_count, size_t npixels, int blocks,
|
|
cudaStream_t stream) {
|
|
T masked;
|
|
if constexpr (std::is_signed_v<T>)
|
|
masked = std::numeric_limits<T>::min();
|
|
else
|
|
masked = std::numeric_limits<T>::max();
|
|
accumulate_kernel<T><<<blocks, 256, 0, stream>>>(reinterpret_cast<const T *>(raw), nframes,
|
|
frame_bytes / sizeof(T), max_value, sum_value,
|
|
valid_count, npixels, masked);
|
|
}
|
|
|
|
} // namespace
|
|
|
|
ShadowAccumulatorGPU::ShadowAccumulatorGPU(size_t npixels)
|
|
: npixels(npixels),
|
|
stream(std::make_shared<CudaStream>()),
|
|
gpu_max(npixels),
|
|
gpu_sum(npixels),
|
|
gpu_count(npixels) {
|
|
cuda_err(cudaMemsetAsync(gpu_max, 0, sizeof(int64_t) * npixels, *stream));
|
|
cuda_err(cudaMemsetAsync(gpu_sum, 0, sizeof(int64_t) * npixels, *stream));
|
|
cuda_err(cudaMemsetAsync(gpu_count, 0, sizeof(uint32_t) * npixels, *stream));
|
|
|
|
int device = 0;
|
|
cuda_err(cudaGetDevice(&device));
|
|
cudaDeviceProp prop{};
|
|
cuda_err(cudaGetDeviceProperties(&prop, device));
|
|
blocks = 8 * prop.multiProcessorCount;
|
|
|
|
// Size the batch for the widest pixel type there is, so nothing has to be allocated once frames
|
|
// start arriving - a cudaMalloc then would stall every worker behind it.
|
|
raw_capacity = npixels * sizeof(uint32_t) * BATCH;
|
|
raw = CudaDevicePtr<uint8_t>(raw_capacity);
|
|
cuda_err(cudaStreamSynchronize(*stream));
|
|
}
|
|
|
|
bool ShadowAccumulatorGPU::Supports(const CompressedImage &image) {
|
|
if (!BSLZ4DecoderGPU::Supports(image))
|
|
return false;
|
|
switch (image.GetMode()) {
|
|
case CompressedImageMode::Int8:
|
|
case CompressedImageMode::Uint8:
|
|
case CompressedImageMode::Int16:
|
|
case CompressedImageMode::Uint16:
|
|
case CompressedImageMode::Int32:
|
|
case CompressedImageMode::Uint32:
|
|
return true;
|
|
default:
|
|
return false;
|
|
}
|
|
}
|
|
|
|
void ShadowAccumulatorGPU::EnsureRawCapacity(size_t bytes_per_frame) {
|
|
if (bytes_per_frame * BATCH > raw_capacity) {
|
|
// The constructor already sized this for the widest pixel type, so in practice this only
|
|
// runs if that guess was too small. cudaMalloc and cudaFree both synchronise the whole
|
|
// device, which is why it is never done per frame.
|
|
FoldPending();
|
|
cuda_err(cudaStreamSynchronize(*stream));
|
|
raw_capacity = bytes_per_frame * BATCH;
|
|
raw = CudaDevicePtr<uint8_t>(raw_capacity);
|
|
}
|
|
// Only once the pending batch has been folded: FoldPending strides the buffer by frame_bytes,
|
|
// so it has to keep describing the frames already in it until they are gone.
|
|
frame_bytes = bytes_per_frame;
|
|
}
|
|
|
|
// Fold the frames decoded so far into the projection. One pass over the accumulator for the whole
|
|
// batch rather than one per frame.
|
|
void ShadowAccumulatorGPU::FoldPending() {
|
|
if (pending == 0)
|
|
return;
|
|
switch (pending_mode) {
|
|
case CompressedImageMode::Int8:
|
|
launch<int8_t>(raw, pending, frame_bytes, gpu_max, gpu_sum, gpu_count, npixels, blocks, *stream); break;
|
|
case CompressedImageMode::Uint8:
|
|
launch<uint8_t>(raw, pending, frame_bytes, gpu_max, gpu_sum, gpu_count, npixels, blocks, *stream); break;
|
|
case CompressedImageMode::Int16:
|
|
launch<int16_t>(raw, pending, frame_bytes, gpu_max, gpu_sum, gpu_count, npixels, blocks, *stream); break;
|
|
case CompressedImageMode::Uint16:
|
|
launch<uint16_t>(raw, pending, frame_bytes, gpu_max, gpu_sum, gpu_count, npixels, blocks, *stream); break;
|
|
case CompressedImageMode::Int32:
|
|
launch<int32_t>(raw, pending, frame_bytes, gpu_max, gpu_sum, gpu_count, npixels, blocks, *stream); break;
|
|
case CompressedImageMode::Uint32:
|
|
launch<uint32_t>(raw, pending, frame_bytes, gpu_max, gpu_sum, gpu_count, npixels, blocks, *stream); break;
|
|
default:
|
|
throw JFJochException(JFJochExceptionCategory::InputParameterInvalid,
|
|
"ShadowAccumulatorGPU: unsupported image mode");
|
|
}
|
|
pending = 0;
|
|
}
|
|
|
|
void ShadowAccumulatorGPU::Add(const CompressedImage &image) {
|
|
if (!Supports(image))
|
|
throw JFJochException(JFJochExceptionCategory::InputParameterInvalid,
|
|
"ShadowAccumulatorGPU: image cannot be decoded on the device");
|
|
if (static_cast<size_t>(image.GetWidth()) * image.GetHeight() != npixels)
|
|
throw JFJochException(JFJochExceptionCategory::InputParameterInvalid,
|
|
"ShadowAccumulatorGPU: image size does not match the detector");
|
|
|
|
// A batch holds one pixel type and one frame size; a change of either closes the batch first,
|
|
// while frame_bytes and pending_mode still describe the frames already in it.
|
|
if (pending > 0 && (image.GetMode() != pending_mode
|
|
|| image.GetUncompressedSize() != frame_bytes))
|
|
FoldPending();
|
|
|
|
if (!decoder || image.GetUncompressedSize() != frame_bytes)
|
|
decoder = std::make_unique<BSLZ4DecoderGPU>(image.GetUncompressedSize(), stream);
|
|
EnsureRawCapacity(image.GetUncompressedSize());
|
|
pending_mode = image.GetMode();
|
|
|
|
decoder->Decode(image, raw.get() + static_cast<size_t>(pending) * frame_bytes);
|
|
// The decode has to be known good before its frame is counted, and the flag only arrives once
|
|
// the stream has drained.
|
|
cuda_err(cudaStreamSynchronize(*stream));
|
|
decoder->ThrowIfDecodeFailed();
|
|
|
|
pending++;
|
|
frames++;
|
|
if (pending == BATCH)
|
|
FoldPending();
|
|
}
|
|
|
|
void ShadowAccumulatorGPU::Download(std::vector<int64_t> &max_value, std::vector<int64_t> &sum_value,
|
|
std::vector<uint32_t> &valid_count) {
|
|
FoldPending();
|
|
max_value.resize(npixels);
|
|
sum_value.resize(npixels);
|
|
valid_count.resize(npixels);
|
|
cuda_err(cudaMemcpyAsync(max_value.data(), gpu_max.get(), sizeof(int64_t) * npixels,
|
|
cudaMemcpyDeviceToHost, *stream));
|
|
cuda_err(cudaMemcpyAsync(sum_value.data(), gpu_sum.get(), sizeof(int64_t) * npixels,
|
|
cudaMemcpyDeviceToHost, *stream));
|
|
cuda_err(cudaMemcpyAsync(valid_count.data(), gpu_count.get(), sizeof(uint32_t) * npixels,
|
|
cudaMemcpyDeviceToHost, *stream));
|
|
cuda_err(cudaStreamSynchronize(*stream));
|
|
}
|